from pathlib import Path
from collections.abc import Callable
import typer
import yaml
import platform
import os

def belongto(oripart:Callable, target:Callable) -> bool:
    for i in oripart :
        if i in target :
            return True
    return False

def setting_config():
    # 处理配置位置
    overpath = os.getenv("PYSSWORDSZ")
    if overpath is None :
        if platform.system() == "Windows" :
            overpath = "AppData/Local/pysswordsz"
        else :
            overpath = ".config/pysswordsz"
        conPath = Path.home() / overpath
    else :
        conPath = overpath
    return conPath

def newConfig() -> None:
    data = {
        "columns": ["url","user","comment"],
        "keyfolder":setting_config(),
        "datafolder":setting_config()
    }
    changefolder = typer.confirm("是否更改密钥KEY文件的保存位置?\nDo you want to change the KEY file location?\n--> ")
    if changefolder :
        data["keyfolder"] = typer.prompt("请输入密钥KEY文件的保存位置\nPlease input the KEY file location\n--> ")
    changefolder = typer.confirm("是否更改数据DATA文件的保存位置?\nDo you want to change the DATA file location?\n--> ")
    if changefolder :
        data["datafolder"] = typer.prompt("请输入数据DATA文件的保存位置\nPlease input the DATA file location\n--> ")
    os.mkdir(setting_config())
    with open(setting_config() / "config.yaml", "w", encoding="utf-8") as fx :
        yaml.dump(data, fx)

class pszconfig(object):
    def __init__(self) -> None:
        self.__home = setting_config() / "config.yaml"
        with open(self.__home, "r", encoding="utf-8") as ftxt:
            self.__data = yaml.load(ftxt.read(),Loader=yaml.FullLoader)
    def __saveconfig(self) -> None :
        with open(self.__home, "w", encoding="utf-8") as fx :
            yaml.dump(self.__data, fx)
    def keyfolder(self) -> Path:
        fp = self.__data["keyfolder"]
        return Path(fp)
    def datafolder(self) -> Path:
        fp = self.__data["datafolder"]
        return Path(fp)
    @property
    def columns(self) -> list:
        return self.__data["columns"]
    @property
    def vaultlist(self) -> list:
        if "vaultlist" in self.__data.keys():
            return self.__data["vaultlist"]
        else :
            return []
    @property
    def vault(self) -> str:
        if "vault" in self.__data.keys():
            return self.__data["vault"]
        else :
            raise ValueError("No vault is set!")
    def list(self) -> None:
        print(self.__data)
    def setting(self, name:str, value:str) -> None:
        if name == "columns":
            if "," in value :
                newcols = value.spilt(",")
            else:
                newcols = list(set(self.columns.append(value)))
            if belongto(newcols,["uuid","name","password","createtime"]) :
                raise KeyError('"uuid" "name" "password" and "createtime" are reserved names.')
            else:
                self.__data[name] = newcols
        else:
            self.__data[name] = value
        self.__saveconfig()
        print("Complete the configuration settings!")
    def remove(self, name:str) :
        if name not in self.__data.keys():
            raise KeyError("The configuration does not exist!")
        if name in ["columns","keyfolder","datafolder"]:
            if name == "columns":
                self.__data[name] = ["url","user","comment"]
            else :
                self.__data[name] = setting_config()
        else :
            self.__data.pop(name)
        self.__saveconfig()
        print("Complete removing config {} !".format(name))


from pathlib import Path
from pysswordsz.pzsconfig import pszconfig
from pysswordsz.encrytool import encryting,generatePassword
from uuid import uuid4
from datetime import datetime

import typer

import polars as pl

def buildPWDB(name:str) -> None:
    cons = pszconfig()
    filePlace = cons.datafolder()
    columns = cons.columns
    data = {
        i:[] for i in (["uuid","name"]+columns+["password","createtime"])
    }
    data = pl.DataFrame(data)
    data.write_csv(filePlace / (name+".lyz"))
    cons.setting("vault",name)
    if cons.vaultlist :
        if name in cons.vaultlist :
            raise ValueError("vault {} already exists".format(name))
        else:
            cons.setting("vaultlist",cons.vaultlist.append(name))
    else:
        cons.setting("vaultlist",[name])
    print("Complete build a vault {}".format(name))

def ask_password() -> list :
    resList = {}
    resList["n"] = typer.prompt("输入密码长度 --> ",default=16,type=int)
    is_url = typer.confirm("是否需要URL安全字符? --> ")
    if is_url :
        resList["urlsafe"] = is_url
    else:
        resList["need_number"] = typer.confirm("是否需要数字? --> ",default=True)
        resList["need_upper"] = typer.confirm("是否需要大写字母? --> ",default=True)
        resList["eed_punctuation"] = typer.confirm("是否需要符号? --> ",default=True)
        resList["mina"] = typer.prompt("各类型字符的最小个数 --> ",default=1,type=int)
    return resList

class pwsmanager(object):
    def __init__(self) -> None:
        cons = pszconfig()
        self.__home = cons.datafoldar()
        self.__default = cons.vault
        self.__vaultList = cons.vaultlist
        self.__cipher = encryting()
        self.__columns = cons.columns
    def __read(self, vault:str) -> pl.DataFrame:
        theVault = self.__default if vault=="default" else vault
        if theVault not in self.__vaultList :
            raise ValueError("vault {} not exists".format(theVault))
        return pl.read_csv(self.__home / (theVault+".lyz"))
    def update(self,name:str, vault:str = "default"):
        theVault = self.__default if vault=="default" else vault
        alldata = self.__read(theVault)
        if name not in alldata["name"].unique().to_list():
            raise ValueError("password {} not exists. You need create one".format(name))
        else:
            olddata = alldata.filter(pl.col("name")==name)
        newdata = {
            "uuid":uuid4(),
            "name":name}
        if typer.confirm("是否需要直接生成新密码？ --> ") :
            password = generatePassword(n=16)
        else :
            password = generatePassword(**ask_password())
        newdata["password"] = self.__cipher.encrypt_data(password)
        for i in self.__columns:
            newdata[i] = olddata.select(pl.col(i)).unique().to_numpy()[0][0]
        newdata["createtime"] = datetime.now()
        alldata = alldata.vstack(pl.DataFrame(newdata))
        alldata.write_csv(self.__home / (theVault+".lyz"))
        print("Complete updated a password of {} ...".format(name))
    def search(self, name:str, all:bool = False, vault:str = "default"):
        data = self.__read(vault)
        if name not in data["name"].unique().to_list():
            raise ValueError("password {} not exists. You need create one".format(name))
        else:
            sdata = data.filter(pl.col("name")==name)
            if sdata.is_empty():
                sdata = data.filter(pl.col("name").str.contains(name))
        if sdata.is_empty():
            print("[red]No result found!")
            return pl.DataFrame()
        else:
            sdata = sdata.with_columns(
                password = pl.col("password").map_elements(lambda x : self.__cipher.decrypt_data(x),return_dtype=pl.String)
            )
            if all:
                return sdata.sort("createtime",descending=True)
            else:
                return sdata.sort("createtime",descending=True).head(1)
    def add_password(self,name:str, to:str = "default"):
        theVault = self.__default if to=="default" else to
        if theVault not in self.__vaultList :
            raise ValueError("vault {} not exists".format(theVault))
        add_Data = {
            "uuid":uuid4(),
            "name":name}
        if typer.confirm("是否需要生成新密码？ --> ") :
            password = generatePassword(**ask_password())
        else :
            password = typer.prompt("请输入密码 --> ",hide_input=True,confirmation_prompt=True)
        add_Data["password"] = self.__cipher.encrypt_data(password)
        for i in self.__columns:
            txt = typer.prompt("please input the {} of {} --> ".format(i,name))
            is_encr = typer.confirm("是否需要加密保存? --> ")
            if is_encr :
                add_Data[i] = self.__cipher.encrypt_data(txt)
            else :
                add_Data[i] = txt
        add_Data["createtime"] = datetime.now()
        alldata = self.__read(to)
        alldata = alldata.vstack(pl.DataFrame(add_Data))
        alldata.write_csv(self.__home / (theVault+".lyz"))
        print("Complete adding a password of {} into {} vault...".format(name, to))
    def delete(self, name:str, vault:str = "default"):
        theVault = self.__default if vault=="default" else vault
        data = self.__read(theVault)
        if name not in data["name"].unique().to_list():
            raise ValueError("password {} not exists.".format(name))
        else:
            data = data.filter(pl.col("name")!=name)
            data.write_csv(self.__home / (theVault+".lyz"))
        print("Complete deleted passwords {} from {} ...".format(name,vault))
    def list(self, vault:str = "default", other_columns:list[str]|None = None):
        alldata = self.__read(vault)
        if other_columns :
            selcols = ["uuid","name"] + other_columns + ["createtime"]
        return alldata.select(pl.col(selcols))


import string
import shutil
import polars as pl
from datetime import datetime
import os
from collections.abc import Callable
from secrets import token_urlsafe,randbelow,choice
from random import shuffle
from math import floor,ceil
from typing import Any
from pysswordsz.pzsconfig import pszconfig
from getpass import getpass
from Crypto.PublicKey import RSA
from Crypto.Random import get_random_bytes,random
from Crypto.Cipher import AES, PKCS1_OAEP
from Crypto.IO import PEM
from zipfile import ZipFile,ZIP_DEFLATED
from uuid import uuid3,NAMESPACE_URL
from pathlib import Path
from typer import prompt
from pysswordsz.pzsconfig import belongto


def hasin(oripart:Callable, target:Any) -> int:
    resn = 0
    for i in oripart :
        if i == target :
            resn = resn + 1
    return resn

def toprandList(top:int, n:int = 4, minz:int = 1) -> list[int] :
    res = []
    loT = floor(top * 2 / 3)
    for i in range(n-1) :
        tx = randbelow(loT)
        if tx < (top / n) and i < 1:
            tx = floor(top / n)
        if tx < minz :
            tx = tx + minz
        res.append(tx)
        loT = floor((top - sum(res)) * 2 / 3)
    if res[0] > (top / 3) :
        res.append(top - sum(res))
    else :
        res = [top - sum(res)] + res
    return res

def generateXKCDPassword(n:int = 6, type_mode:str = "pinyin",
                         padding:bool = False) -> str :
    if type_mode not in ["pinyin","wubi","english","german","spanish","mix"]:
        raise ValueError("type_mode {} not exists".format(type_mode))
    file = Path(os.path.dirname(__file__)) / "static/{}.wordlist".format(type_mode)
    sep = " " if padding is False else "!@$%^&*-_+=:|~?/.;"
    coln = 0 if type_mode in ["pinyin","wubi"] else 1
    words = pl.read_csv(file,has_header=False,separator="\t")
    wordlist = [words[randbelow(words.height)].to_numpy()[0,coln].capitalize() for i in range(n)]
    if padding :
        xn = ceil(n/4)
        seqList = ["".join(random.sample(sep,xn)) for i in range(n)]
        pw = "".join([seqList[i]+wordlist[i] for i in range(n)]) + "".join(random.sample(sep,xn))
    else:
        pw = sep.join(wordlist)
    return pw

def generatePassword(n:int, need_number:bool = True, 
                     need_upper:bool = True, need_punctuation:bool = True, 
                     mina:int = 1, urlsafe:bool = False) -> str :
    if urlsafe :
        n_x = floor(n / 1.3)
        pw = token_urlsafe(n_x)
        if not belongto(pw, "_-^@()[]") :
            pw = pw.replace(pw[randbelow(n)], "_")
    else :
        partnum = hasin([need_number,need_punctuation,need_upper], True)
        basechars = toprandList(top=n,n=(partnum+1),minz=mina)
        keyChar = [string.ascii_lowercase]
        if need_number :
            keyChar.append(string.digits)
        if need_upper :
            keyChar.append(string.ascii_uppercase)
        if need_punctuation :
            keyChar.append("!~@#$%^&*()_+=-[]|:;?<>,.{}")
        res = []
        for i in range(partnum+1) :
            res = res + [choice(keyChar[i]) for jj in range(basechars[i])]
        shuffle(res)
        shuffle(res)
        pw = "".join(res)
    return pw

def newKeys() -> None:
    print("输入你的主密码。\n但请特别注意：务必记住主密码，一旦丢失主密码，则加密信息和保存的密码将不再能打开！")
    passwordx = prompt("Core Password:", hide_input=True, confirmation_prompt=True)
    print("Waiting for creating the keys...")
    private = RSA.generate(3072)
    keyHome = pszconfig().keyfolder()
    with open(keyHome/"theKey.lyz", "wb") as f:
        data = private.export_key(passphrase=passwordx,
                                pkcs=8,
                                protection='PBKDF2WithHMAC-SHA512AndAES256-CBC',
                                prot_params={'iteration_count':131072})
        f.write(data)
    print("The keys have been created!")
    print("!!Reminder!!\n\tDon't forget your Core Password!")

class encryting(object):
    def __init__(self) -> None:
        passwordx = getpass("Core Password :")
        keyHome = pszconfig().keyfolder()
        self.__dataHome = pszconfig().datafolder()/"encinfoDB.lyz"
        with open(keyHome/"theKey.lyz", "rb") as f:
            data = f.read()
        self.__private = RSA.import_key(data, passphrase=passwordx)
        self.__public = self.__private.public_key()
    def encrypt_data(self, data:str) -> str:
        cipher = PKCS1_OAEP.new(self.__public)
        ciphertext = cipher.encrypt(data.encode("utf-8"))
        return PEM.encode(ciphertext, "DATA")
    def decrypt_data(self, data:str) -> str:
        odata = PEM.decode(data)[0]
        cipher = PKCS1_OAEP.new(self.__private)
        plaintext = cipher.decrypt(odata)
        return plaintext.decode()
    def encrypt_file(self, file:str|Path, comments:str|None = None) -> None:
        oriPath = Path(file)
        if oriPath.is_dir() :
            zipPath = oriPath.parent/("{}.zip".format(oriPath.name))
            with ZipFile(zipPath, 
                         compression = ZIP_DEFLATED, compresslevel = 7,
                         mode = "w") as fz :
                for i in oriPath.glob("*") :
                    fz.write(i, compress_type=ZIP_DEFLATED, compresslevel=7)
            shutil.rmtree(oriPath)
        with open((zipPath if oriPath.is_dir() else oriPath), "rb") as f:
            data = f.read()
        session_key = get_random_bytes(16)
        cipher_rsa = PKCS1_OAEP.new(self.__public)
        enc_session_key = cipher_rsa.encrypt(session_key)
        cipher_aes = AES.new(session_key, AES.MODE_EAX)
        ciphertext, tag = cipher_aes.encrypt_and_digest(data)
        save_name = oriPath.parent/("{}.lyz".format(oriPath.name))
        with open(save_name, "wb") as f:
            f.write(enc_session_key)
            f.write(cipher_aes.nonce)
            f.write(tag)
            f.write(ciphertext)
        thisData = {
             "UUID": [uuid3(NAMESPACE_URL,str(oriPath))],
             "name": [("{}.lyz".format(oriPath.name))],
             "path": [oriPath.parent],
             "type": ["file" if oriPath.is_file() else "folder"],
             "comments": [comments],
             "time": [datetime.now().strftime("%Y-%m-%d %H:%M:%S")]
        }
        encrList = self.list_files()
        encrList = encrList.vstack(pl.from_dicts(thisData))
        encrList.write_csv(self.__dataHome)
        shutil.rmtree((zipPath if oriPath.is_dir() else oriPath))
    def decrypt_file(self, file:str|Path) -> None:
        with open(Path(file), "rb") as f:
            enc_session_key = f.read(self.__private.size_in_bytes())
            nonce = f.read(16)
            tag = f.read(16)
            ciphertext = f.read()
        cipher_rsa = PKCS1_OAEP.new(self.__private)
        session_key = cipher_rsa.decrypt(enc_session_key)
        cipher_aes = AES.new(session_key, AES.MODE_EAX, nonce)
        data = cipher_aes.decrypt_and_verify(ciphertext, tag)
        disname = Path(file).name.replace(".lyz", "")
        if "." in disname :
            save_name = Path(file).parent/disname
        else :
            save_name = Path(file).parent/(disname+".zip")
        with open(save_name, "wb") as f:
            f.write(data)
        shutil.rmtree(Path(file))
        with ZipFile(save_name, "r") as fz :
            fz.extractall(Path(file).parent)
        shutil.rmtree(save_name)
        encrList = self.list_files()
        encrList = encrList.filter(
            pl.col("UUID") != uuid3(NAMESPACE_URL,str(Path(file).parent/disname)))
        encrList.write_csv(self.__dataHome)
    def list_files(self) -> pl.DataFrame:
        if self.__dataHome.exists() :
            return pl.read_csv(self.__dataHome)
        else :
            return pl.DataFrame()





import typer
import pyperclip as clip
import polars as pl

from typing_extensions import Annotated
from pysswordsz.encrytool import generatePassword,newKeys,encryting,generateXKCDPassword
from pysswordsz.pzsconfig import newConfig,pszconfig
from rich import print
from pysswordsz.pwsmanager import pwsmanager,buildPWDB

app = typer.Typer(no_args_is_help=True)

config = typer.Typer()
app.add_typer(config, name="config", help="manage config \t\t 管理配置")
passdb = typer.Typer()
app.add_typer(passdb, name="pass", help="manage password \t 管理密码")
cryptl = typer.Typer()
app.add_typer(cryptl, name="crypt", help="encrypt or decrypt \t 加密解密")

@app.command("version", help="show version \t\t 显示版本")
def version():
    print("VERSION 0.1.0")
    print("pysswordSz Copyright (C) 2024  Sidney Zhang <zly@lyzhang.me>")
    print("Licensed under GPL-3.0 license.")

@app.command(help="generate password \t 生成密码")
def genpass(n:Annotated[int,typer.Argument(help="密码位数，建议普通密码至少10位，xkcd密码至少8词",show_default=False)] = 16,
            xkcd:Annotated[bool,typer.Argument(help="生成XKCD密码")] = False,
            need_upper:Annotated[bool,typer.Argument(help="需要大写字母")] = True,
            need_number:Annotated[bool,typer.Argument(help="需要数字")] = True,
            need_punctuation:Annotated[bool,typer.Argument(help="需要序号")] = True,
            minia:Annotated[int,typer.Argument(help="各类字符的最小个数")] = 1,
            urlsafe:Annotated[bool,typer.Argument(help="是否生成链接安全密码")] = False,
            xk_mode:Annotated[str,typer.Argument(help="xkcd密码生成模式")] = "pinyin",
            xk_padding:Annotated[bool,typer.Argument(help="xkcd密码中是否使用随机填充")] = False,
            show:Annotated[bool,typer.Argument(help="是否展示密码")] = True,
            to_clip:Annotated[bool,typer.Argument(help="是否复制到系统粘贴板")] = False):
    if xkcd :
        pswds = generateXKCDPassword(n=n, type_mode=xk_mode,
                                     padding=xk_padding)
    else:
        pswds = generatePassword(n = n, need_punctuation=need_punctuation,
                                 need_number=need_number,need_upper=need_upper,
                                 mina=minia, urlsafe=urlsafe)
    if show :
        print(pswds)
    if to_clip :
        clip.copy(pswds)

@app.command(help="init pysswordSz \t 初始化pysswordSz")
def init():
    newConfig()
    newKeys()
    print("The initialization of pysswordSz is complete!")

@config.command("list", help="list all configs / 列出所有已有配置")
def cfg_list():
    data = pszconfig().list()
    print(data)

@config.command("rm", help="remove one of config / 删除一个配置")
def cfg_remove(name:Annotated[str,typer.Argument(help="需要移除的配置名")]):
    pszconfig().remove(name=name)


@config.command("set", help="set one of config / 设置一个配置")
def cfg_setting(name:Annotated[str,typer.Argument(help="需要添加的配置名")],
                value:Annotated[str,typer.Argument(help="配置值")]):
    pszconfig().setting(name=name,value=value)

@passdb.command("build",help="build a vault / 建立密码库")
def pss_build():
    buildPWDB()

@passdb.command("add", help="add a password to vault / 向密码库添加密码")
def pss_add(name:Annotated[str,typer.Argument(help="需要添加的密码名称")],
            to:Annotated[str,typer.Argument(help="需要添加密码的密码库，默认为最后创建的库",show_default=False)]="default"):
    pwsmanager().add_password(name = name, to = to)

@passdb.command("search",help="search a password in vault / 在密码库中搜索密码")
def pss_search(name:Annotated[str,typer.Argument(help="要查找的密码名称")], 
               last_more:Annotated[int,typer.Argument(help="展示旧密码，-1将展示全部，0为不展示，其他正整数为展示长度。")] = 0, 
               vault:Annotated[str,typer.Argument(help="需要查找的密码库",show_default=False)]="default",
               show:Annotated[bool,typer.Argument(help="是否展示密码")] = False, 
               to_clip:Annotated[bool,typer.Argument(help="是否把最新的密码复制到系统粘贴板")] = True):
    mgr = pwsmanager()
    need_all = False if last_more == 0 else True
    data = mgr.search(name = name, all=need_all, vault=vault)
    if data.is_empty():
        print("")
    else:
        if need_all :
            with pl.Config(tbl_rows=-1):
                hl = last_more if last_more>0 else data.height
                needPWS = data.head(hl).select(pl.col("uuid","password","createtime"))
                dddd = needPWS["password"].to_numpy()[0]
                if show:
                    print(needPWS)
                else:
                    print(needPWS.filter(pl.col("password")!=dddd))
                if to_clip:
                    clip.copy(dddd)
        else:
            keyname = data.columns
            for i in keyname:
                if i !="password":
                    print("{} :\t{}".format(i,data[i][0]))
                else:
                    if show :
                        print("{} :\t{}".format(i,data[i][0]))
                    if to_clip :
                        clip.copy(data[i][0])

@passdb.command("update", help="update a password in vault / 更新密码库中的密码")
def pss_update(name:Annotated[str,typer.Argument(help="要更新的密码名称")], 
               vault:Annotated[str,typer.Argument(help="指定密码库",show_default=False)]="default"):
    pwsmanager().update(name=name, vault=vault)

@passdb.command("delete",help="delete a password in vault / 删除密码库中的密码")
def pss_delete(name:Annotated[str,typer.Argument(help="要删除的密码名称")], 
               vault:Annotated[str,typer.Argument(help="指定密码库",show_default=False)]="default"):
    pwsmanager().delete(name=name, vault=vault)

@passdb.command("list", help="list all passwords in vault / 列出密码库中的所有密码")
def pss_list(vault:Annotated[str,typer.Argument(help="指定所需密码库",show_default=False)]="default", 
             add_columns:Annotated[str,typer.Argument(help="增加需要展示的列名，使用`,`隔开。",show_default=False)] = ""):
    addcolumns = add_columns.split(",") if add_columns else None
    showall = typer.confirm("是否显示所有信息? --> ")
    needrows = -1 if showall else 16
    data = pwsmanager().list(vault=vault, other_columns=addcolumns)
    if data.is_empty():
        print("[green]There is currently no password in Vault {}...".format(vault))
    else :
        with pl.Config(tbl_rows=needrows):
            print(data)

@cryptl.command("encr",help="encrypt a file / 加密一个文件或文件夹")
def ctl_encr(file: Annotated[str, typer.Argument(help="需要加密的文件或文件夹路径")]):
    cipher = encryting()
    cipher.encrypt_file(file)
    print("[green]Encryption is complete!")

@cryptl.command("decr",help="decrypt a file / 解密一个文件或文件夹")
def ctl_decr(file: Annotated[str, typer.Argument(help="需要解密的文件路径")]):
    cipher = encryting()
    cipher.decrypt_file(file)
    print("[green]Decryption is complete!")

@cryptl.command("list",help="list all files / 列出所有加密文件")
def ctl_list():
    dataList = encryting().list_files()
    if dataList.is_empty():
        print("There is currently no encrypted data...")
    else:
        with pl.Config(tbl_rows=-1):
            print(dataList)

def entry_point() -> None:
    app()

def main() -> None:
    entry_point()

if __name__ == "__main__":
    entry_point()